/**
  *  \file planetscentral/dbclient/dbkey.cc
  *  \brief Database Client: Key
  */

#include "planetscentral/dbclient/dbkey.h"
#include "int/data.h"
#include "planetscentral/dbclient/dbclient.h"

/** Constructor.
    \param client Database connection
    \param s Name of key */
DbKey::DbKey(DbClient& client, const string_t& s)
    : client(client),
      key(s)
{ }

DbKey::~DbKey()
{ }

/** Check for existence.
    \retval true Key exists
    \retval false Key does not exist */
bool
DbKey::exists()
{
    return client.callInt(DbRequest().withString("EXISTS").withString(key)) != 0;
}

/** Remove this key from the database. */
void
DbKey::remove()
{
    client.callVoid(DbRequest().withString("DEL").withString(key));
}

/** Rename this key. Overwrites %newName if it already exists.
    \param newName new name for this key. The DbKey object will change as well. */
void
DbKey::rename(string_t newName)
{
    client.callVoid(DbRequest().withString("RENAME").withString(key).withString(newName));
    this->key = newName;
}

/** Rename this key. Does not overwrite %newName if it already exists.
    \param newName new name for this key. The DbKey object will change as well.
    \retval true rename successful
    \retval false target already exists */
bool
DbKey::renameNX(string_t newName)
{
    bool result = client.callInt(DbRequest().withString("RENAMENX").withString(key).withString(newName));
    if (result) {
        this->key = newName;
    }
    return result;
}

/** Get type of this key. */
DbKey::Type
DbKey::getType()
{
    string_t result = client.callString(DbRequest().withString("TYPE").withString(key));
    if (result == "none") {
        return None;
    } else if (result == "string") {
        return String;
    } else if (result == "list") {
        return List;
    } else if (result == "set") {
        return Set;
    } else if (result == "zset") {
        return ZSet;
    } else if (result == "hash") {
        return Hash;
    } else {
        return Unknown;
    }
}

/** Get name of this key. */
const string_t&
DbKey::getName() const
{
    return key;
}

/** Get database client used for this key. */
DbClient&
DbKey::getClient()
{
    return client;
}
