/**
  *  \file planetscentral/dbclient/dbzset.h
  *  \brief Database Client: ZSets
  */
#ifndef C2SERVER_PLANETSCENTRAL_DBCLIENT_DBZSET_H
#define C2SERVER_PLANETSCENTRAL_DBCLIENT_DBZSET_H

#include "planetscentral/dbclient/dbkey.h"
#include "planetscentral/dbclient/connection.h"
#include "planetscentral/dbclient/dbsort.h"

class DbZSetSlice;

/** ZSet key. A ZSet contains elements ("fields") sorted by a value ("score").
    Operations include manipulating the scores, and extracting ranges of the
    sorted sequence.

    Whereas redis allows the scores to be double-precision floating point, we
    only support integers, which should be enough for us. */
class DbZSet : public DbKey {
 public:
    enum Direction {
        Ascending,
        Descending
    };
    DbZSet(DbClient& client, string_t name);

    /* ZADD key score value    int, 0 (updated), 1 (added)             zset[key][value] := score */
    bool setField(string_t name, int32_t value);
    /* ZCARD key               int                                     zset[key].size() */
    int32_t size();
    /* ZINCRBY key score val   double, new score                       zset[key][value] += score */
    int32_t addToField(string_t name, int32_t value);
    /* ZRANK key value         int-or-nil                              position of value in zset, ascending score */
    /* ZREVRANK key value      int-or-nil                              position of value in zset, descending score */
    int32_t indexOf(string_t name, Direction dir);
    /* ZREM key value          int, 0 (not there), 1 (removed)         zset[key].erase(value) */
    bool removeField(string_t name);
    /* ZSCORE key value        double-or-nil                           zset[key][value] */
    int32_t get(string_t name);
    bool hasField(string_t name);
    /* ZRANGE k s e [WITHSCORES]  list<string>, values [and scores]    slice of sorted zset, ascending score */
    /* ZREVRANGE k s e [WITHSCORES]  list<string>, vals [and scores]   slice of sorted zset, deascending score */
    void getRange(int32_t start, int32_t end, Direction dir, StringList_t& out);
    /* ZREMRANGEBYRANK k s e   int, #elements removed                  ZREM k (ZRANGE k s e) */
    int32_t removeRange(int32_t start, int32_t end);

    DbZSetSlice valueRange(int32_t low, int32_t high);

    DbSort sort();
};

class DbZSetSlice {
 public:
    DbZSetSlice(DbZSet zset, int32_t low, int32_t high);

    /* ZCOUNT k i x            int                                     size of slice of sorted zset, score i..x */
    int32_t size();
    /* ZREMRANGEBYSCORE k i x  int, #elements removed                  ZREM k (ZRANGEBYSCORE k i x) */
    int32_t remove();

    /* ZRANGEBYSCORE k i x [LIMIT o c] [WITHSCORES]  list<string>...   slice of sorted zset, score i..x */
    void get(StringList_t& out);
    void getRange(int32_t start, int32_t end, StringList_t& out);

 private:
    DbZSet zset;
    int32_t low;
    int32_t high;
};


/* ZINTERSTORE [...]       int, #elements in new set               intersect N zsets, combine scores */
/* ZUNIONSTORE [...]       int, #elements in new set               merge N zsets, combine scores */

#endif
