/**
  *  \file planetscentral/mailin/mailin.cc
  *  \brief Mailin: Main Program
  *
  *  This is the main program of the "mailin" service, which takes
  *  a mail message and attempts to interpret it for the host service.
  */

#include <cstring>
#include <stdexcept>
#include <stdio.h>
#include "cpluslib/format.h"
#include "int/array.h"
#include "planetscentral/dbclient/connection.h"
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/net/mimeparser.h"
#include "server/config.h"
#include "cpluslib/buffer.h"
#include "io/streams.h"


/**************************** Debugging: Dump ****************************/

static bool
dumpWantHeader(const string_t& name)
{
    return name == "subject"
        || name == "from"
        || name == "to"
        || name == "date"
        || name == "message-id"
        || name == "TRACE"
        || (name.size() > 8 && name.compare(0, 8, "content-", 8) == 0);
}

static void
dump(MimeParser& p, int indent)
{
    // Headers
    const MimeParser::HeaderMap_t& h = p.getHeaders();
    for (MimeParser::HeaderMap_t::const_iterator i = h.begin(); i != h.end(); ++i) {
        if (dumpWantHeader(i->first)) {
            printf("%*s%s: %s\n", indent, "", i->first.c_str(), i->second.c_str());
        }
    }
    string_t fn = p.getFileName();
    if (!fn.empty()) {
        printf("%*sFileName: %s\n", indent, "", fn.c_str());
    }
    printf("\n");

    // Content
    MimeMultipartParser mpp;
    if (mpp.init(p)) {
        MimeParser part;
        int i = 0;
        while (mpp.readPart(part)) {
            printf("%*s---- Part %d:\n", indent, "", ++i);
            dump(part, indent + 5);
        }
        if (i == 0) {
            printf("%*s---- Empty Multi-Part Message\n", indent, "");
        }
    } else {
        string_t ctyp = strLCase(p.getHeader("content-type"));
        if (ctyp.empty() || (ctyp.size() > 4 && strLCase(ctyp.substr(0, 4)) == "text")) {
            const MimeParser::BodyVec_t& b = p.getBody();
            for (MimeParser::BodyVec_t::const_iterator i = b.begin(); i != b.end(); ++i) {
                string_t line = *i;
                if (line.size() > 75) {
                    line.erase(70);
                    line += "...";
                }
                printf("%*s%s\n", indent, "", line.c_str());
            }
        } else {
            printf("%*s(Non-Text Content)\n", indent, "");
        }
    }
}

/*************************** Regular Operation ***************************/

static string_t host_host    = "127.0.0.1";
static string_t host_port    = "7777";
static DbClient host_conn;
static string_t mailout_host = "127.0.0.1";
static string_t mailout_port = "21212";
static DbClient mailout_conn;
static string_t arg_rejects;

static string_t
quote(string_t s)
{
    for (string_t::size_type i = s.size(); i > 0; --i) {
        if (s[i-1] == '\\' || s[i-1] == '"') {
            s.insert(i-1, "\\");
        }
    }
    return s;
}

static void
processConfig(string_t key, string_t value)
{
    if (key == "HOST.HOST") {
        host_host = value;
    } else if (key == "HOST.PORT") {
        host_port = value;
    } else if (key == "MAILOUT.HOST") {
        mailout_host = value;
    } else if (key == "MAILOUT.PORT") {
        mailout_port = value;
    } else if (key == "MAILIN.REJECTDIR") {
        arg_rejects = value;
    }
}

static void
sendRejection(const MimeParser& root, const string_t& address, const string_t& path, const char* tpl)
{
    printf("[mismatch] file '%s'\n", path.c_str());

    // Connect to mailer:
    if (!mailout_conn.isConnected()) {
        mailout_conn.connectRetry(mailout_host, mailout_port);
    }

    // Send mail
    // Do not use a uniquifier; this goes to a mail address, not a user, and thus is not queued
    mailout_conn.callVoid(DbRequest().withString("MAIL").withString(tpl));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_from").withString(address));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_subject").withString(root.getHeader("subject")));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_date").withString(root.getHeader("date")));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_messageid").withString(root.getHeader("message-id")));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_path").withString(path));
    mailout_conn.callVoid(DbRequest().withString("SEND").withString("mail:" + address));
}

static IntValue*
lookup(IntValue* a, const char* key)
{
    IntArray* av = dynamic_cast<IntArray*>(a);
    if (!av) {
        return 0;
    }
    Ptr<IntArrayData> ad = av->getData();
    for (uint32_t i = 0; i+1 < ad->content.size(); i += 2) {
        if (IntValue::toStringOf(ad->content[i], false) == key) {
            return ad->content[i+1];
        }
    }
    return 0;
}

static bool
processTurnFile(MimeParser& root, const string_t& content, string_t address, string_t path)
{
    // Connect to host
    if (!host_conn.isConnected()) {
        host_conn.connectRetry(host_host, host_port);
    }

    // Find user agent for info string
    // X-Mailer: used by Outlook, Eudora, Pegasus, The Bat!, Lotus Notes, phpBB3, ...
    // User-Agent: used by Mozilla, Mutt, KMail, Opera Mail, Alpine, ...
    string_t ua = root.getHeader("x-mailer");
    if (ua.empty()) {
        ua = root.getHeader("user-agent");
    }
    
    // Submit a turn file command. TRN...MAIL requires admin privileges.
    std::auto_ptr<IntValue> result;
    try {
        host_conn.callVoid(DbRequest().withString("USER").withString(string_t()));
        result.reset(host_conn.call(DbRequest()
                                    .withString("TRN")
                                    .withString(content)
                                    .withString("MAIL")
                                    .withString(address)
                                    .withString("INFO")
                                    .withString("mail: addr=\"" + quote(address)
                                                + "\", ua=\"" + quote(ua)
                                                + "\", route=\"" + quote(root.getHeader("TRACE")) + "\"")));
    }
    catch (std::runtime_error& e) {
        if (std::strncmp(e.what(), "407 ", 4) == 0) {
            // "407 Mail mismatch" happens when the player does not play in this game
            sendRejection(root, address, path, "turn-mismatch");
            return true;
        } else if (std::strncmp(e.what(), "404 ", 4) == 0 || std::strncmp(e.what(), "412 ", 4) == 0) {
            // "412 Wrong game state" happens when trying to submit a turn to a finished game
            // "404 Game does not exist" happens when then turn file does not match any game
            sendRejection(root, address, path, "turn-stale");
            return true;
        } else {
            // Anything else, e.g. internal errors. Don't bother uses with those.
            throw;
        }
    }

    // Generate a reply
    //   status    (0=missing, 1=green, 2=yellow, 3=red, 4=bad)
    //   previous  (0=missing, 1=green, 2=yellow, 3=red, 4=bad)
    //   output    (turnchecker output)
    //   game      (game number)
    //   slot      (player slot)
    //   user      (user Id)
    int32_t status    = toIntResult(lookup(result.get(), "status"));
    int32_t slot      = toIntResult(lookup(result.get(), "slot"));
    string_t game     = toStringResult(lookup(result.get(), "game"));
    string_t output   = toStringResult(lookup(result.get(), "output"));
    string_t user     = toStringResult(lookup(result.get(), "user"));
    string_t gameName = host_conn.callString(DbRequest().withString("GAMEGETNAME").withString(game));
    int32_t gameTurn  = host_conn.callInt(DbRequest().withString("GAMEGET").withString(game).withString("turn"));
    printf("[ok] file '%s': user '%s', game '%s', slot %d, state %d\n", path.c_str(), user.c_str(), game.c_str(), int(slot), int(status));

    // Connect to mailer:
    if (!mailout_conn.isConnected()) {
        mailout_conn.connectRetry(mailout_host, mailout_port);
    }

    // Send mail
    mailout_conn.callVoid(DbRequest().withString("MAIL").withString("turn").withString(format("turn-%s-%s-%d", user, game, slot)));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("gamename").withString(gameName));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("gameid").withString(game));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("gameturn").withInt(gameTurn));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("slot").withInt(slot));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("trn_status").withInt(status));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("trn_output").withString(output));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_subject").withString(root.getHeader("subject")));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_date").withString(root.getHeader("date")));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_messageid").withString(root.getHeader("message-id")));
    mailout_conn.callVoid(DbRequest().withString("PARAM").withString("mail_path").withString(path));
    mailout_conn.callVoid(DbRequest().withString("SEND").withString("user:" + user));

    return true;
}

static bool
processMailPart(MimeParser& root, MimeParser& part, string_t address, string_t path)
{
    // Is it multipart?
    MimeMultipartParser mpp;
    if (mpp.init(part)) {
        // Process multipart
        int i = 0;
        bool result = false;
        MimeParser subpart;
        while (mpp.readPart(subpart)) {
            ++i;
            result |= processMailPart(root, subpart, address, format("%s/part%d", path, i));
        }
        return result;
    } else {
        // Try to process single part
        string_t fn = strLCase(part.getFileName());
        if (fn.size() > 4 && fn.compare(fn.size()-4, 4, ".trn", 4) == 0) {
            return processTurnFile(root, part.getBodyAsString(), address, path + "/" + fn);
        }
        return false;
    }
}

static void
logHeader(MimeParser& p, const char* name)
{
    const MimeParser::HeaderMap_t& h = p.getHeaders();
    MimeParser::HeaderMap_t::const_iterator it(h.find(name));
    if (it != h.end()) {
        printf("  %s: %s\n", strUCFirst(name).c_str(), it->second.c_str());
    }
}

static void
saveReject(Buffer<char>& b)
{
    // Do we want to save rejects?
    if (arg_rejects.empty()) {
        return;
    }

    // Generate file name
    time_t now = time(0);
    char timestamp[50];
    strftime(timestamp, sizeof(timestamp), "%Y%m%d-%H%M%S", gmtime(&now));
    int index = 0;
    string_t name;
    do {
        name = format("%s/%s-%d", arg_rejects, timestamp, ++index);
    } while (Stream::openFileNT(name, Stream::C_READ));

    // Save
    try {
        Ptr<Stream> s = Stream::openFile(name, Stream::C_CREATE);
        s->writeT(b.ptr(), b.getSize());
        printf("[reject] saved as '%s'\n", name.c_str());
    }
    catch (std::exception& e) {
        printf("[error] %s, while writing '%s'\n", e.what(), name.c_str());
    }
}

static void
processMail(Buffer<char>& b)
{
    // Load configuration
    loadConfiguration(processConfig);

    // Parse mail
    MimeParser p;
    p.addData(b.ptr(), b.getSize());
    p.finish();

    // Log the mail
    printf("Processing mail message:\n");
    logHeader(p, "from");
    logHeader(p, "to");
    logHeader(p, "date");
    logHeader(p, "message-id");

    // Parse the "From" address
    string_t address = getHeaderAddressValue(p.getHeader("from"));
    if (address.empty()) {
        printf("[reject] unable to figure out sender address\n");
        saveReject(b);
    } else {
        try {
            if (processMailPart(p, p, address, string_t())) {
                printf("[ok] message processed\n");
            } else {
                printf("[reject] no usable content in message\n");
                saveReject(b);
            }
        }
        catch (std::exception& e) {
            printf("[reject] [exception] %s\n", e.what());
            saveReject(b);
        }
    }
}

int main(int /*argc*/, char** argv)
{
    // Parse args
    const char* arg_prog = *argv;
    bool opt_dump = false;
    while (const char* p = *++argv) {
        if (p[0] == '-' && p[1] == '-') {
            ++p;
        }
        if (std::strcmp(p, "-dump") == 0) {
            opt_dump = true;
        } else if (std::strcmp(p, "-help") == 0) {
            printf("usage: %s [-dump] < MAILFILE\n", arg_prog);
            return 0;
        } else {
            fprintf(stderr, "%s: invalid option/argument.\n", arg_prog, p);
            return 1;
        }
    }

    // Read mail from stdin
    Buffer<char> b;
    char buffer[1024];
    int n;
    while ((n = fread(buffer, 1, sizeof(buffer), stdin)) > 0) {
        b.append(buffer, n);
    }

    // Process it
    if (opt_dump) {
        MimeParser p;
        p.addData(b.ptr(), b.getSize());
        p.finish();
        dump(p, 0);
    } else {
        processMail(b);
    }
}
