/**
  *  \file planetscentral/mailout/message.cc
  *  \brief Mailout: Message
  */

#include "planetscentral/mailout/message.h"
#include "planetscentral/mailout/transmitter.h"   // getCurrentTime

/** Constructor.
    \param root  Subtree of the "mqueue:" database
    \param mid   Message Id
    \param state State of this message ("preparing", "sending") */
Message::Message(DbSubtree root, int32_t mid, string_t state)
    : root(root),
      msg(root.subtree("msg").subtree(mid)),
      mid(mid),
      state(state)
{ }

/** Name of template used for this message. */
DbHashStringField
Message::templateName()
{
    return msg.hashKey("data").stringField("template");
}

/** Unique Id of this message. */
DbHashStringField
Message::uniqueId()
{
    return msg.hashKey("data").stringField("uniqid");
}

/** Hash with all message parameters. */
DbHash
Message::arguments()
{
    return msg.hashKey("args");
}

/** List with all message attachment URLs. */
DbStringList
Message::attachments()
{
    return msg.stringListKey("attach");
}

/** Set of message receivers. Each receiver is a pair of addressing scheme and
    address, as in "user:<uid>" or "mail:<user>@<host>". */
DbStringSet
Message::receivers()
{
    return msg.stringSetKey("to");
}

/** Timestamp when message was sent. */
DbHashIntField
Message::expireTime()
{
    return msg.hashKey("data").intField("expire");
}

/** Delete this message. It will not be sent. */
void
Message::remove()
{
    msg.hashKey("data").remove();
    arguments().remove();
    attachments().remove();
    receivers().remove();
    root.intSetKey(state).remove(mid);
}

/** Send this message. Moves the message into the sending queue, and makes sure
    that no other message with the same uniqueId() will be sent.

    Note that this only updates the database copy of the send queue; it does NOT
    place the message in the in-memory send queue of the transmitter; you must
    call transmitter.send(getId()) after calling this function. */
void
Message::send()
{
    string_t uniqid = uniqueId().get();
    if (!uniqid.empty()) {
        root.hashKey("uniqid").setField(uniqid, mid);
    }
    if (state != "sending") {
        root.intSetKey(state).moveTo(mid, root.intSetKey("sending"));
        state = "sending";
    }
}

/** Allocate a new message. Use this when sending new mail. */
std::auto_ptr<Message>
allocateMessage(DbSubtree root)
{
    int32_t mid = ++root.subtree("msg").intKey("id");
    root.intSetKey("preparing").add(mid);

    std::auto_ptr<Message> result(new Message(root, mid, "preparing"));
    result->expireTime().set(getCurrentTime() + getMaximumMessageAge());
    return result;
}
