/**
  *  \file planetscentral/nntp/nntpworker.cc
  *  \brief NNTP Server Instance
  *
  *  This serves one NNTP connection.
  *  The main thread creates instances of this class, and hands them
  *  to the worker thread, which transfers data into and out of it.
  *
  *  Minimum command set:
  *  - AUTHINFO (required for authentication)
  *  - QUIT (trivial)
  *  - LIST ACTIVE (required to get list of groups)
  *  - GROUP (required to enter a group)
  *  - ARTICLE/HEAD (required to access articles)
  *  This is enough to make 'tin' work.
  *
  *  Additional commands to make it practical:
  *  - HELP (trivial)
  *  - BODY/STAT (trivial once we have ARTICLE)
  *  - OVER/XOVER (Mozilla doesn't work without)
  *  - LIST SUBSCRIPTIONS (why not)
  *  - LIST OVERVIEW.FMT (required for XOVER)
  *  - MODE (trivial, ignored)
  *
  *  References:
  *  - RFC 3977 Network News Transfer Protocol (NNTP)
  *  - RFC 2980 Common NNTP Extensions
  *  - hamsrv (an earlier NNTP implementation I wrote)
  */

#include <algorithm>
#include <vector>
#include <stdexcept>
#include <stdio.h>
#include <cstring>
#include "planetscentral/nntp/nntpworker.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "int/array.h"
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/nntp/nntp.h"
#include "planetscentral/osutil.h"
#include "util/misc.h"

static const char TOO_FEW_ARGS[]        = "501 Too few arguments\r\n";
static const char TOO_MANY_ARGS[]       = "501 Too many arguments\r\n";
static const char SYNTAX_ERROR[]        = "501 Syntax error\r\n";
static const char NOT_SUPPORTED_MAJOR[] = "500 Unsupported command\r\n";
static const char NOT_SUPPORTED_MINOR[] = "501 Unsupported command\r\n";
static const char NEED_AUTH[]           = "480 Need authentication\r\n";
static const char NOT_IN_GROUP[]        = "412 Not currently in a newsgroup\r\n";

/** List of overview fields. */
static const char*const OVERVIEW_FIELDS[] = {
    "Subject",
    "From",
    "Date",
    "Message-ID",
    "References",
    ":bytes",
    ":lines",
    "Xref",
};

/** Index of first "full" overview field. All but the standard fields must be "full"
    fields including the header name. Therefore, this value follows directly from
    the specification.

    'tin' only accepts the Xref header in full format.
    If Xref is not listed in full format, it attempts to access Xref using XHDR
    which we do not support. */
static const size_t OVERVIEW_FIELDS_FIRST_FULL = 7;

/** Eat a word from the string.
    \param cmd [in/out] String
    \return First word of string */
string_t
eatWord(string_t& cmd)
{
    string_t::size_type i = cmd.find_first_not_of(" \t\r\n");
    string_t result;
    if (i == string_t::npos) {
        cmd.clear();
    } else {
        string_t::size_type n = cmd.find_first_of(" \t\r\n", i);
        if (n == string_t::npos) {
            result.assign(cmd, i, cmd.size()-i);
            cmd.clear();
        } else {
            result.assign(cmd, i, n-i);
            cmd.erase(0, n);
        }
    }
    return result;
}

/** Eat up remainder of string.
    \param cmd [in/out] String (will be cleared)
    \return Rest of string */
string_t
eatRest(string_t& cmd)
{
    string_t result(strTrim(cmd));
    cmd.clear();
    return result;
}

/** Sanitize a header field value.
    Header fields can contain spaces, tabs, and newlines, which cannot be transmitted
    in a news overview file. This function sanitizes them by replacing runs of
    \r\n\t or space, starting with a \r\n\t, by a single space. */
string_t
sanitizeFieldValue(string_t value)
{
    string_t::size_type i = 0;
    while (1) {
        string_t::size_type j = value.find_first_of("\r\n\t", i);
        if (j == string_t::npos) {
            break;
        }
        string_t::size_type k = value.find_first_not_of("\r\n\t ", j);
        if (k == string_t::npos) {
            value.erase(j);
            break;
        }
        value.replace(j, k-j, 1, ' ');
        i = j+1;
    }
    return value;
}

/** Escape dots.
    Prepends dots to lines starting with a dot. */
string_t
escapeDots(string_t value)
{
    // Start of string is the start of a line
    string_t::size_type i = 0;
    while (i < value.size()) {
        if (value[i] == '.') {
            value.insert(i, 1, '.');
        }

        // Find next line. Since linefeeds are CRLF, it suffices to look for LF.
        i = value.find('\n', i);
        if (i == string_t::npos) {
            break;
        }
        ++i;
    }
    return value;
}

/******************************* NntpWorker ******************************/

/** Constructor. */
NntpWorker::NntpWorker()
    : receive_buffer(),
      send_buffer(),
      status(ReadCommand),
      auth_status(NeedUser),
      auth_user(),
      auth_uid(0),
      grouplist_cache(),
      current_group(),
      current_forum(0),
      current_seq(0),
      current_seq_map()
{
    send("200 c2nntp says hello\r\n");
}

NntpWorker::~NntpWorker()
{ }

void
NntpWorker::handleRead(const char* data, std::size_t size)
{
    const std::size_t LIMIT = 20000;
    while (size > 0) {
        const char* p = std::strchr(data, '\n');
        if (p == 0) {
            if (receive_buffer.size() < LIMIT) {
                receive_buffer.append(data, size);
            }
            break;
        } else {
            std::size_t amount = p - data;
            if (receive_buffer.size() < LIMIT) {
                receive_buffer.append(data, amount);
                try {
                    switch (status) {
                     case ReadCommand:  handleCommand(receive_buffer);  break;
                     case ReadPostData: handlePostData(receive_buffer); break;
                     case Terminated:                                   break;
                    }
                }
                catch (std::exception& e) {
                    MutexGuard g(output_mutex);
                    printf("[exception] %s\n", e.what());
                    send("403 Internal error\r\n");
                }
            } else {
                send("501 Line too long\r\n");
            }
            data += amount+1;
            size -= amount+1;
            receive_buffer.clear();
        }
    }
}

void
NntpWorker::handleWrite(std::size_t size)
{
    send_buffer.erase(0, std::min(size, send_buffer.size()));
}

std::size_t
NntpWorker::getBytesToSend(const char*& data)
{
    if (send_buffer.size() > 0) {
        data = send_buffer.data();
        return send_buffer.size();
    } else {
        return 0;
    }
}

bool
NntpWorker::isTerminated()
{
    return status == Terminated && send_buffer.empty();
}

void
NntpWorker::markEndOfInput()
{
    status = Terminated;
}

/** Schedule some data for sending.
    \param s Data to send */
void
NntpWorker::send(string_t s)
{
    send_buffer += s;
}

/** Schedule some response data for sending.
    In addition to scheduling the data for sending, logs it.
    \param s Data to send */
void
NntpWorker::sendResponse(string_t s)
{
    {
        MutexGuard g(output_mutex);
        string_t::size_type n = s.find_first_of("\r\n");
        if (n == s.npos) {
            n = s.size();
        }
        printf("[id:%d] < %.*s\n", getId(), n, s.data());
    }

    send(s);
}

/** Handle NNTP command.
    \param cmd Command */
void
NntpWorker::handleCommand(string_t cmd)
{
    // Figure out command verb
    string_t verb = strUCase(eatWord(cmd));
    if (verb.empty()) {
        send(SYNTAX_ERROR);
        return;
    }

    // Log it
    if (verb != "AUTHINFO") {
        MutexGuard g(output_mutex);
        printf("[id:%d] > %s", getId(), verb.c_str());
        if (!cmd.empty()) {
            printf("%s", cmd.c_str());
        }
        printf("\n");
    }

    // Process it
    if (verb == "QUIT") {
        sendResponse("205 Good bye\r\n");
        status = Terminated;
    } else if (verb == "ARTICLE") {
        // ARTICLE (977, 3977)
        handleArticle(cmd, true, true);
    } else if (verb == "AUTHINFO") {
        // AUTHINFO (2980)
        handleAuthinfo(cmd);
    } else if (verb == "BODY") {
        // BODY (977, 3977)
        handleArticle(cmd, false, true);
    } else if (verb == "GROUP") {
        // GROUP (977, 3977)
        handleGroup(cmd);
    } else if (verb == "HEAD") {
        // HEAD (977, 3977)
        handleArticle(cmd, true, false);
    } else if (verb == "HELP") {
        // HELP (977, 3977)
        handleHelp();
    } else if (verb == "LIST") {
        // LIST (977, 2980, 3977)
        handleList(cmd);
    } else if (verb == "LISTGROUP") {
        // LISTGROUP (2980, 3977)
        handleListGroup(cmd);
    } else if (verb == "MODE") {
        // MODE STREAM|READER (2980, 3977)
        sendResponse("200 Ignored\r\n");
    } else if (verb == "STAT") {
        // STAT (977, 3977)
        handleArticle(cmd, false, false);
    } else if (verb == "OVER" || verb == "XOVER") {
        handleOver(cmd);
    } else {
        // CAPABILITIES                                                    3977
        // CHECK <msgid>                                                   2980
        // DATE                                                            2980, 3977
        //          Answer: yyyymmddhhmmss
        // HDR                                                             3977
        //          see XHDR
        // IHAVE <msgid>                                                   977, 3977
        // LAST                                                            977, 3977
        //          decrease internal current-article pointer
        //          Answer: sequence number, msgid
        // NEWGROUPS yymmdd hhmmss [GMT] [distributions]                   977, 3977
        //          Answer: list of newsgroups
        // NEWNEWS newsgroups date time [GMT] [distributions]              977, 3977
        // NEXT                                                            977, 3977
        //          see LAST
        // POST                                                            977, hamsrv, 3977
        // SLAVE                                                           977
        // TAKETHIS <msgid>                                                2980
        // XGTITLE [wildmat]       same as LIST NEWSGROUPS, deprecated     2980
        // XHDR header [range|<msgid>]                                     2980
        //          Answer: header field value for an article
        // XINDEX [wildmat]        deprecated                              2980
        // XPAT header range|<msgid> pat...                                2980
        // XPATH <msgid>           deprecated                              2980
        // XREPLIC ...                                                     2980
        // XROVER [range]                                                  2980
        //          Same as XHDR References
        // XTHREAD [DBINIT|THREAD] deprecated                              2980
        sendResponse(NOT_SUPPORTED_MAJOR);
    }
}

void
NntpWorker::handlePostData(const string_t& line)
{
    (void) line;
}

/** Check authentication.
    \retval true Authentication succeeded, command processing can proceed
    \retval false Not authenticated. An error message has been sent, command processing must abort */
bool
NntpWorker::checkAuth()
{
    if (auth_status != Authenticated) {
        sendResponse(NEED_AUTH);
        return false;
    } else {
        talk_connection.callVoid(DbRequest().withString("USER").withInt(auth_uid));
        return true;
    }
}

/** Fill cache containing newsgroups.
    \retval true Cache is available, proceed
    \retval false Error. An error message has been sent, command processing must abort */
bool
NntpWorker::fillGroupListCache()
{
    // Boilerplate
    ReconnectingGuard g(talk_rc);
    if (!checkAuth()) {
        return false;
    }

    // Do we have it already?
    if (grouplist_cache.get() != 0) {
        return true;
    }

    // Reload
    grouplist_cache.reset(talk_connection.call(DbRequest().withString("NNTPLIST")));
    return true;
}

/** Resolve sequence number into message number.
    \param seq [in] Sequence number
    \param rfcMsgId [out] Message Id
    \retval nonzero Message number
    \retval 0 Error. An error message has been sent, command processing must abort */
int32_t
NntpWorker::resolveSequenceNumber(int32_t seq, string_t& rfcMsgId)
{
    // Must be in a forum
    if (current_forum == 0) {
        sendResponse(NOT_IN_GROUP);
        return 0;
    }

    // Look up in cache
    std::map<int32_t, int32_t>::const_iterator it = current_seq_map.find(seq);
    if (it == current_seq_map.end() || it->second == 0) {
        sendResponse("423 No such article number\r\n");
        return 0;
    }

    // Get message Id
    rfcMsgId = talk_connection.callString(DbRequest().withString("POSTGET").withInt(it->second).withString("rfcmsgid"));

    // OK
    return it->second;
}

/** Parse sequence number range.
    \param [in] range Range specified by user
    \param [out] min Lower bound
    \param [out] max Upper bound
    \retval true Success
    \retval false Error. An error message has been sent, command processing must abort */
bool
NntpWorker::parseRange(const string_t& range, int32_t& min, int32_t& max)
{
    if (range.empty()) {
        min = max = current_seq;
        return true;
    } else {
        string_t::size_type i = range.find('-');
        int32_t a;
        if (i == string_t::npos) {
            if (stringValue(range, a)) {
                min = max = a;
                return true;
            } else {
                sendResponse(SYNTAX_ERROR);
                return false;
            }
        } else {
            // Minimum
            if (i == 0) {
                min = 1;
            } else if (stringValue(range.substr(0, i), a)) {
                min = a;
            } else {
                sendResponse(SYNTAX_ERROR);
                return false;
            }

            // Maximum
            if (i+1 == range.size()) {
                max = 0x7FFFFFFF;
            } else if (stringValue(range.substr(i+1), a)) {
                max = a;
            } else {
                sendResponse(SYNTAX_ERROR);
                return false;
            }

            return true;
        }
    }
}

/** Enter a newsgroup.
    This loads the cache of sequence numbers to message numbers.
    \param groupName Name of newsgroup
    \retval true Success
    \retval false Error. An error message has been sent, command processing must abort */
bool
NntpWorker::enterGroup(const string_t& groupName)
{
    // Find newsgroup. FIXME: use grouplist_cache?
    std::auto_ptr<IntValue> groupInfo;
    try {
        groupInfo.reset(talk_connection.call(DbRequest().withString("NNTPFINDNG").withString(groupName)));
    }
    catch (...) {
        sendResponse("411 No such group\r\n");
        return false;
    }

    // Parse it
    IntArray* groupInfoA = dynamic_cast<IntArray*>(groupInfo.get());
    if (!groupInfoA) {
        throw std::runtime_error("Didn't get a list");
    }
    int32_t forumId = 0;
    IntDataSegment& groupData = groupInfoA->getData()->content;
    for (uint32_t i = 0, n = groupData.size(); i+1 < n; i += 2) {
        string_t key = toStringResult(groupData[i]);
        if (key == "id") {
            forumId = toIntResult(groupData[i+1]);
        }
    }

    // OK?
    if (forumId == 0) {
        sendResponse("411 No such group\r\n");
        return false;
    }

    // OK, group exists. Load list of sequence numbers.
    current_group = groupName;
    current_forum = forumId;
    current_seq = 0;
    current_seq_map.clear();

    std::auto_ptr<IntValue> seqMap(talk_connection.call(DbRequest().withString("NNTPFORUMLS").withInt(forumId)));
    IntList_t seqList;
    toIntListResult(seqMap.get(), seqList);

    for (IntList_t::size_type i = 0, n = seqList.size(); i+1 < n; i += 2) {
        current_seq_map.insert(std::make_pair(seqList[i], seqList[i+1]));
    }

    return true;
}


/*
 *  Commands
 */


/** ARTICLE/HEAD/BODY/STAT command.
    - References: RFC 977, RFC 3977, hamsrv
    - Syntax: <verb> msgid
    - Syntax: <verb> sequenceNumber

    Responses:
    - 220 n message-id Article follows (multi-line)
    - 221 n message-id Headers follow (multi-line)
    - 222 n message-id Body follows (multi-line)
    - 223 n message-id Article exists
    - 412 No newsgroup selected
    - 423 No article with that number
    - 420 Current article number is invalid (is this possible?)

    \param header true to send the header (HEAD, ARTICLE)
    \param body true to send the body (BODY, ARTICLE) */
void
NntpWorker::handleArticle(string_t args, bool header, bool body)
{
    // Boilerplate
    ReconnectingGuard g(talk_rc);
    if (!checkAuth()) {
        return;
    }

    // Check Id
    string_t id(eatWord(args));
    int32_t msgId;
    int32_t seqNr;
    string_t rfcMsgId;
    if (id == "") {
        // No parameter: get current article
        seqNr = current_seq;
        msgId = resolveSequenceNumber(current_seq, rfcMsgId);
        if (msgId == 0) {
            // Message sent by resolveSequenceNumber
            return;
        }
    } else if (stringValue(id, seqNr)) {
        // Sequence number
        msgId = resolveSequenceNumber(seqNr, rfcMsgId);
        if (msgId == 0) {
            // Message sent by resolveSequenceNumber
            return;
        }
    } else if (id.size() > 2 && id[0] == '<' && id[id.size()-1] == '>') {
        // could be a mid
        try {
            seqNr = 0;
            rfcMsgId = id.substr(1, id.size()-2);
            msgId = talk_connection.callInt(DbRequest().withString("NNTPFINDMID").withString(rfcMsgId));
        }
        catch (...) {
            sendResponse("430 No such article\r\n");
            return;
        }
    } else {
        sendResponse(SYNTAX_ERROR);
        return;
    }

    // At this point, msgId is a resolved message number.
    // Fetch required data.
    // If anything fails here, we'll get into the
    // throw-an-exception-log-internal-error branch,
    // which should be good enough.

    Ptr<IntArrayData> headerArray;
    string_t bodyString;
    if (header) {
        std::auto_ptr<IntValue> rawHead(talk_connection.call(DbRequest().withString("NNTPPOSTHEAD").withInt(msgId)));
        if (IntArray* a = dynamic_cast<IntArray*>(rawHead.get())) {
            headerArray = a->getData();
        }
        if (headerArray == 0) {
            throw std::runtime_error("Did not get an array");
        }
    }
    if (body) {
        bodyString = escapeDots(talk_connection.callString(DbRequest()
                                                           .withString("POSTRENDER")
                                                           .withInt(msgId)
                                                           .withString("FORMAT")
                                                           .withString("news")
                                                           .withString("BASEURL")
                                                           .withString(www_prefix)));
    }

    // Send success response
    int code = (header
                ? body ? 220 : 221
                : body ? 222 : 223);
    sendResponse(format("%d %d <%s> found\r\n", code, seqNr, rfcMsgId));
    if (header) {
        for (uint32_t i = 0, n = headerArray->content.size(); i+1 < n; i += 2) {
            string_t headerName(toStringResult(headerArray->content[i]));
            if (headerName.size() > 0 && headerName[0] != ':') {
                send(headerName);
                send(": ");
                send(toStringResult(headerArray->content[i+1]));
                send("\r\n");
            }
        }
    }
    if (header && body) {
        send("\r\n");
    }
    if (body) {
        send(bodyString);
        send("\r\n");
    }
    if (header || body) {
        send(".\r\n");
    }
}

/** AUTHINFO command.
    - References: RFC 2980
    - Syntax: AUTHINFO USER user
    - Syntax: AUTHINFO PASS pass
    - Syntax: AUTHINFO SIMPLE
    - Syntax: AUTHINFO GENERIC args

    Responses:
    - 281 Authentication accepted
    - 381 More authentication information required
    - 480 Authentication required
    - 482 Authentication rejected
    - 502 No permission

    USER/PASS is the simplest scheme and is implemented here.
    tin uses AUTHINFO GENERIC as a first guess, passing us the log-in name.
    That would probably work in a trusted setup with identd,
    but there's no way to make use of that for our application. */
void
NntpWorker::handleAuthinfo(string_t cmd)
{
    // Do we have a sub-verb? Also log it.
    string_t kind(strUCase(eatWord(cmd)));
    {
        MutexGuard g(output_mutex);
        printf("[id:%d] > AUTHINFO %s [...]\n", getId(), kind.c_str());
    }
    if (kind.empty()) {
        sendResponse(TOO_FEW_ARGS);
        return;
    }

    if (kind == "USER") {
        // AUTHINFO USER
        auth_user = eatRest(cmd);
        auth_status = NeedPass;
        sendResponse("381 Send password\r\n");
    } else if (kind == "PASS") {
        // AUTHINFO PASS
        if (auth_status == NeedPass) {
            ReconnectingGuard g(talk_rc);
            try {
                auth_uid = talk_connection.callInt(DbRequest().withString("NNTPUSER").withString(auth_user).withString(eatRest(cmd)));
                auth_status = Authenticated;
                grouplist_cache.reset();
                sendResponse("281 Authentification accepted\r\n");

                MutexGuard g(output_mutex);
                printf("[id:%d] [user:%d] Authenticated as '%s'\n", getId(), auth_uid, auth_user.c_str());
            }
            catch (...) {
                sendResponse("482 Authentification rejected\r\n");
                auth_status = NeedUser;
            }
        } else {
            sendResponse("501 Need AUTHINFO USER first\r\n");
        }
    } else {
        sendResponse(NOT_SUPPORTED_MINOR);
    }
}

/** GROUP command.
    - References: RFC 977, RFC 3977, hamsrv
    - Indicating capability: READER
    - Syntax: GROUP groupname

    Responses:
    - 211 num-articles first last sequence groupname
    - 411 No such newsgroup */
void
NntpWorker::handleGroup(string_t args)
{
    // Must be authenticated
    ReconnectingGuard g(talk_rc);
    if (!checkAuth()) {
        return;
    }

    // Must have group name
    string_t groupName(eatRest(args));
    if (groupName.empty()) {
        sendResponse(SYNTAX_ERROR);
        return;
    }

    // Look up group
    if (!enterGroup(groupName)) {
        return;
    }

    // Generate output
    if (current_seq_map.empty()) {
        sendResponse(format("211 0 0 0 %s is empty\r\n", groupName));
    } else {
        int32_t firstSeq = current_seq_map.begin()->first;
        int32_t lastSeq = (--current_seq_map.end())->first;
        sendResponse(format("211 %d %d %d %s selected\r\n", lastSeq - firstSeq + 1, firstSeq, lastSeq, groupName));
        current_seq = firstSeq;
    }
}

/** LIST command.
    Dispatches to various sub-commands. */
void
NntpWorker::handleList(string_t args)
{
    // LIST ACTIVE.TIMES                                               2980, 3977
    //          Answer: newsgroup creation-time creator-email
    // LIST DISTRIBUTIONS                                              2980
    // LIST DISTRIB.PATS                                               2980, 3977
    // LIST HEADERS                                                    3977
    string_t what = strUCase(eatWord(args));
    if (what == "" || what == "ACTIVE") {
        handleListActive();
    } else if (what == "NEWSGROUPS") {
        handleListNewsgroups();
    } else if (what == "SUBSCRIPTIONS") {
        handleListSubscriptions();
    } else if (what == "OVERVIEW.FMT") {
        handleListOverviewFormat();
    } else {
        sendResponse(NOT_SUPPORTED_MINOR);
    }
}

/** LIST / LIST ACTIVE command.
    - References: RFC 977, RFC 3977, RFC 2980, hamsrv
    - Indicating capability: READER
    - Syntax: LIST
    - Syntax: LIST ACTIVE [wildmat]

    Responses:
    - 215 (multiline, followed by "newsgroup lastSeq firstSeq postingAllowedFlag")

    FIXME: wildmat is not implemented. */
void
NntpWorker::handleListActive()
{
    // Fetch group list
    if (!fillGroupListCache()) {
        return;
    }

    IntArray* a = dynamic_cast<IntArray*>(grouplist_cache.get());
    if (a == 0) {
        throw std::runtime_error("Didn't get a list");
    }

    // Send it
    sendResponse("215 List of newsgroups follows\r\n");
    IntDataSegment& data = a->getData()->content;
    for (uint32_t i = 0, n = data.size(); i < n; ++i) {
        if (IntArray* ele = dynamic_cast<IntArray*>(data[i])) {
            string_t newsgroup = "error";
            int32_t lastSeq = 0;
            int32_t firstSeq = 1;
            bool postingFlag = false;
            IntDataSegment& eleData = ele->getData()->content;
            for (uint32_t j = 0, n = eleData.size(); j+1 < n; j += 2) {
                string_t key = toStringResult(eleData[j]);
                if (key == "newsgroup") {
                    newsgroup = toStringResult(eleData[j+1]);
                } else if (key == "lastSeq") {
                    lastSeq = toIntResult(eleData[j+1]);
                } else if (key == "firstSeq") {
                    firstSeq = toIntResult(eleData[j+1]);
                } else if (key == "writeAllowed") {
                    postingFlag = toIntResult(eleData[j+1]) != 0;
                }
            }
            send(format("%s %d %d %c\r\n", newsgroup, lastSeq, firstSeq, postingFlag ? 'y' : 'n'));
        }
    }
    sendResponse(".\r\n");
}

/** LIST NEWSGROUPS command.
    - References: RFC 2980, RFC 3977, hamsrv
    - Indicating capability: READER
    - Syntax: LIST NEWSGROUPS [wildmat]

    Responses:
    - 215 information follows (multiline, followed by newsgroup/description pairs)
    - 503 program error, function not performed

    FIXME: wildmat is not implemented. */
void
NntpWorker::handleListNewsgroups()
{
    // Fetch group list
    if (!fillGroupListCache()) {
        return;
    }

    IntArray* a = dynamic_cast<IntArray*>(grouplist_cache.get());
    if (a == 0) {
        throw std::runtime_error("Didn't get a list");
    }

    // Send it
    sendResponse("215 List of newsgroups follows\r\n");
    IntDataSegment& data = a->getData()->content;
    for (uint32_t i = 0, n = data.size(); i < n; ++i) {
        if (IntArray* ele = dynamic_cast<IntArray*>(data[i])) {
            string_t newsgroup = "error";
            string_t description = "";
            IntDataSegment& eleData = ele->getData()->content;
            for (uint32_t j = 0, n = eleData.size(); j+1 < n; j += 2) {
                string_t key = toStringResult(eleData[j]);
                if (key == "newsgroup") {
                    newsgroup = toStringResult(eleData[j+1]);
                } else if (key == "description") {
                    description = toStringResult(eleData[j+1]);
                }
            }

            string_t::size_type n = description.find_first_of("\r\n");
            if (n != string_t::npos) {
                description.erase(n);
            }

            send(format("%s %s\r\n", newsgroup, description));
        }
    }
    sendResponse(".\r\n");
}

/** LIST SUBSCRIPTIONS command.
    - References: RFC 2980
    - Syntax: LIST SUBSCRIPTIONS

    Response codes:
    - 215 information follows (multi-line, list of newsgroups follows)
    - 503 program error, function not performed */
void
NntpWorker::handleListSubscriptions()
{
    // Boilerplate
    ReconnectingGuard g(talk_rc);
    if (!checkAuth()) {
        return;
    }

    // Fetch list
    // For now, the subscription list is defined as root group.
    // We'd have to deviate from that when introducing a special NNTP-only group.
    std::auto_ptr<IntValue> result(talk_connection.call(DbRequest().withString("NNTPGROUPLS").withString("root")));
    StringList_t list;
    toStringListResult(result.get(), list);

    // Send result
    sendResponse("215 Recommendations follow\r\n");
    for (StringList_t::const_iterator i = list.begin(); i != list.end(); ++i) {
        send(*i + "\r\n");
    }
    sendResponse(".\r\n");
}

/** LIST OVERVIEW.FMT command.
    - References: RFC 3977, RFC 2980, hamsrv
    - Indicating capability: OVER
    - Syntax: LIST OVERVIEW.FMT

    Responses:
    - 215 Information follows (multi-line)
    - 503 program error, function not performed */
void
NntpWorker::handleListOverviewFormat()
{
    send("215 List follows\r\n"
         "Subject:\r\n"
         "From:\r\n"
         "Date:\r\n"
         "Message-ID:\r\n"
         "References:\r\n"
         ":bytes\r\n"
         ":lines\r\n"
         "Xref:full\r\n"
         ".\r\n");
}

/** LISTGROUP command.
    - References: RFC 3977, RFC 2980, hamsrv
    - Indicating capability: READER
    - Syntax: LISTGROUP [group [range]]

    Responses:
    - 211 number low high group Article numbers follow (multi-line)
    - 411 No such newsgroup
    - 412 No newsgroup selected

    Actions: same as GROUP (select the group, reset current article),
    and list sequence numbers. */
void
NntpWorker::handleListGroup(string_t args)
{
    // Boilerplate
    ReconnectingGuard g(talk_rc);
    if (!checkAuth()) {
        return;
    }

    // If group specified, enter it
    string_t groupName(eatWord(args));
    if (!groupName.empty()) {
        if (!enterGroup(groupName)) {
            return;
        }
    }
    if (current_forum == 0) {
        sendResponse("412 No newsgroup selected\r\n");
        return;
    }

    // Range specified?
    int32_t min, max;
    string_t range(eatWord(args));
    if (!range.empty()) {
        if (!parseRange(range, min, max)) {
            return;
        }
    } else {
        min = 1;
        max = 0x7FFFFFFF;
    }

    // List
    if (current_seq_map.empty()) {
        sendResponse(format("211 0 0 0 %s is empty\r\n", current_group.c_str()));
    } else {
        int32_t firstSeq = current_seq_map.begin()->first;
        int32_t lastSeq = (--current_seq_map.end())->first;
        sendResponse(format("211 %d %d %d %s selected\r\n", lastSeq - firstSeq + 1, firstSeq, lastSeq, current_group.c_str()));
        current_seq = firstSeq;

        std::map<int32_t, int32_t>::iterator i = current_seq_map.lower_bound(min), e = current_seq_map.end();
        while (i != e && i->first <= max) {
            send(format("%d\r\n", i->first));
            ++i;
        }
    }
    sendResponse(".\r\n");
}

/** HELP command.
    - References: RFC 977, RFC 3977
    - Syntax: HELP

    Responses:
    - 100 help follows (multi-line) */
void
NntpWorker::handleHelp()
{
    send("100 Help\r\n"
         "c2nntp implements the following commands:\r\n"
         "  AUTHINFO {USER user|PASS pass}\r\n"
         "  QUIT\r\n"
         ".\r\n");
}

/** OVER / XOVER command.
    - References: RFC 3977, RFC 2980, hamsrv
    - Syntax: OVER [range]
    - Syntax: OVER <msgid>
    - Syntax: XOVER [range]

    Responses:
    - 224 Overview information follows (multi-line)
    - 430 No article with that message-id
    - 412 No newsgroup selected
    - 423 No articles in that range
    - 420 Current article number is invalid

    The response contains a list of header fields: sequence, Subject, From, Date, Message-Id,
    References, byte size, line count, and optional fields (Xref).

    FIXME: OVER <msgid> is not implemented.

    FIXME: special case for OVER with empty range not implemented (should give 423 instead of 224). */
void
NntpWorker::handleOver(string_t args)
{
    if (auth_status != Authenticated) {
        sendResponse(NEED_AUTH);
        return;
    }

    // Must be in a forum
    if (current_forum == 0) {
        sendResponse(NOT_IN_GROUP);
        return;
    }

    // Resolve range
    int32_t min, max;
    if (!parseRange(eatWord(args), min, max)) {
        return;
    }

    // Build the request
    DbRequest req;
    req.withString("NNTPPOSTMHEAD");
    std::map<int32_t, int32_t>::iterator i = current_seq_map.lower_bound(min), e = current_seq_map.end();
    while (i != e && i->first <= max) {
        req.withInt(i->second);
        ++i;
    }

    // Do it
    std::auto_ptr<IntValue> result(talk_connection.call(req));
    sendResponse("224 Overview follows\r\n");
    if (IntArray* a = dynamic_cast<IntArray*>(result.get())) {
        const IntDataSegment& dseg = a->getData()->content;
        for (std::size_t i = 0; i < dseg.size(); ++i) {
            if (IntArray* ele = dynamic_cast<IntArray*>(dseg[i])) {
                const IntDataSegment& eleSeg = ele->getData()->content;
                string_t values[countof(OVERVIEW_FIELDS)];
                int32_t articleNumber = 0;
                for (std::size_t j = 0; j+1 < eleSeg.size(); j += 2) {
                    string_t fieldName = toStringResult(eleSeg[j]);
                    if (strCaseCmp(fieldName, ":seq") == 0) {
                        articleNumber = toIntResult(eleSeg[j+1]);
                    } else {
                        for (std::size_t fieldIndex = 0; fieldIndex < countof(OVERVIEW_FIELDS); ++fieldIndex) {
                            if (strCaseCmp(fieldName, OVERVIEW_FIELDS[fieldIndex]) == 0) {
                                values[fieldIndex] = sanitizeFieldValue(toStringResult(eleSeg[j+1]));
                                break;
                            }
                        }
                    }
                }
                if (articleNumber != 0) {
                    string_t line = numToString(articleNumber);
                    for (std::size_t fieldIndex = 0; fieldIndex < countof(OVERVIEW_FIELDS); ++fieldIndex) {
                        line += "\t";
                        if (fieldIndex >= OVERVIEW_FIELDS_FIRST_FULL) {
                            line += OVERVIEW_FIELDS[fieldIndex];
                            line += ": ";
                        }
                        line += values[fieldIndex];
                    }
                    line += "\r\n";
                    send(line);
                }
            }
        }
    }
    sendResponse(".\r\n");
}
