/**
  *  \file planetscentral/nntp/nntpworker.h
  */
#ifndef C2SERVER_PLANETSCENTRAL_NNTP_NNTPWORKER_H
#define C2SERVER_PLANETSCENTRAL_NNTP_NNTPWORKER_H

#include <cstddef>
#include <memory>
#include <map>
#include "planetscentral/nntp/worker.h"
#include "cpluslib/string.h"
#include "int/value.h"

class NntpWorker : public Worker {
 public:
    NntpWorker();
    ~NntpWorker();

    virtual void handleRead(const char* data, std::size_t size);
    virtual void handleWrite(std::size_t size);
    virtual std::size_t getBytesToSend(const char*& data);
    virtual bool isTerminated();
    virtual void markEndOfInput();

 private:
    /* Receive buffer. Contains one partial line. */
    string_t receive_buffer;

    /* Send buffer. Contains everything that still has to be sent. */
    string_t send_buffer;

    /* Input Parser Status */
    enum Status {
        ReadCommand,
        ReadPostData,
        Terminated
    };
    Status status;

    /* Authentification */
    enum AuthStatus {
        NeedUser,
        NeedPass,
        Authenticated
    };
    AuthStatus auth_status;                     /**< Status. */
    string_t auth_user;                         /**< User name. */
    int32_t auth_uid;                           /**< User Id. */

    /* Group list cache. Caches the list of newsgroups.
       Since this is not expected to change, it is cached during the whole lifetime of the connection. */
    std::auto_ptr<IntValue> grouplist_cache;

    /* Group status. We cache the sequence->message mappings.
       Since these can change often, we update these whenever a newsgroup is selected,
       and do not try to optimize. */
    string_t current_group;                     /**< Newsgroup name. */
    int32_t current_forum;                      /**< Forum number (fid). */
    int32_t current_seq;                        /**< Current sequence number. */
    std::map<int32_t,int32_t> current_seq_map;  /**< Maps sequence numbers to message numbers (mid). */

    /* Internal primitives */
    void send(string_t s);
    void sendResponse(string_t s);
    void handleCommand(string_t cmd);
    void handlePostData(const string_t& line);
    bool checkAuth();
    bool fillGroupListCache();
    int32_t resolveSequenceNumber(int32_t seq, string_t& rfcMsgId);
    bool parseRange(const string_t& range, int32_t& min, int32_t& max);
    bool enterGroup(const string_t& groupName);

    /* Commands */
    void handleArticle(string_t args, bool header, bool body);
    void handleAuthinfo(string_t args);
    void handleGroup(string_t args);
    void handleList(string_t args);
    void handleListActive();
    void handleListNewsgroups();
    void handleListSubscriptions();
    void handleListOverviewFormat();
    void handleListGroup(string_t args);
    void handleHelp();
    void handleOver(string_t args);
};

#endif
