/**
  *  \file planetscentral/talk/cmdnntp.cc
  *  \brief Talk: NNTP commands
  */

#include <cctype>
#include "planetscentral/talk/cmdnntp.h"
#include "planetscentral/talk/talkconnection.h"
#include "int/args.h"
#include "planetscentral/net/md5.h"
#include "planetscentral/net/encoding.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/dbclient/dbint.h"
#include "planetscentral/dbclient/dbstring.h"
#include "planetscentral/dbclient/dbclient.h"
#include "cpluslib/format.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "int/array.h"
#include "util/misc.h"
#include "planetscentral/talk/forum.h"
#include "planetscentral/talk/message.h"
#include "planetscentral/dbclient/dbsort.h"
#include "planetscentral/talk/cmdpost.h"
#include "planetscentral/talk/group.h"

string_t
simplifyUserName(const string_t str)
{
    // Perl equivalent:
    // $username = lc($username);
    // $username =~ s/[^a-zA-Z0-9]/_/g;
    // $username =~ s/__+/_/g;
    // $username =~ s/^_+//;
    // $username =~ s/_+$//;
    // FIXME: do we already have this elsewhere?

    // Convert/transliterate
    using std::tolower;
    string_t result;
    bool hadSpace = true;
    for (string_t::size_type i = 0, n = str.size(); i < n; ++i) {
        char c = tolower(str[i]);
        if ((c >= 'a' && c <= 'z') || (c >= '0' && c <= '9')) {
            result += c;
            hadSpace = false;
        } else {
            if (!hadSpace) {
                result += '_';
                hadSpace = true;
            }
        }
    }

    // If it ends with illegal characters, hadSpace is set. Remove the final "_".
    if (hadSpace && result.size() > 0) {
        result.erase(result.size()-1);
    }

    // Return
    return result;
}

/* @q NNTPUSER user:UserName pass:Str (Talk Command)
   Authenticate user.
   This checks the user's credentials against the database.
   It does not change the user context.

   Permissions: none.

   @err 401 Failed (invalid user name)
   @err 401 Bad user Id (user does not exist)
   @err 401 Bad hash (invalid user profile)
   @err 401 Bad password (invalid password)

   @retval UID user Id

   @uses uid:$USERNAME, user:$UID:password */
void
doNntpUser(TalkConnection& conn, IntArgBlock& args)
{
    // Check args
    args.checkArgumentCount(2);
    string_t username = simplifyUserName(toStringResult(args.getNext()));
    string_t password = toStringResult(args.getNext());

    // Connect database
    ReconnectingGuard g(database_rc);

    // Valid?
    if (username.empty()) {
        conn.sendError("401 Failed");
        return;
    }

    // Look up user Id
    int32_t uid = DbInt(database_connection, "uid:" + username).get();
    if (uid == 0) {
        conn.sendError("401 Bad user Id");
        return;
    }

    // Get their password
    string_t correctHash = DbString(database_connection, format("user:%d:password", uid)).get();
    if (correctHash.empty()) {
        conn.sendError("401 Bad hash");
        return;
    }

    // Validate
    unsigned char result[16];
    Md5Context ctx;
    ctx.init();
    ctx.update((unsigned char*) user_key.data(), user_key.size());
    ctx.update((unsigned char*) password.data(), password.size());
    ctx.final(result);

    string_t userHash = "1," + encodeBase64(string_t((char*) result, sizeof(result)));
    while (userHash.size() > 0 && userHash[userHash.size()-1] == '=') {
        userHash.erase(userHash.size()-1);
    }
    if (userHash == correctHash) {
        IntIntIntValue iv(uid);
        conn.sendObject(&iv);
    } else {
        conn.sendError("401 Bad password");
    }
}

/* @q NNTPLIST (Talk Command)
   List forums as newsgroups.
   Lists all forums that can be read by the current user.

   Permissions: user context required.

   @retval TalkNewsgroupInfo[] list of newsgroup information
   @uses forum:newsgroups, forum:$FID:header
   @see GROUPLS */
void
doNntpList(TalkConnection& conn, IntArgBlock& args)
{
    // Check args
    args.checkArgumentCount(0);
    conn.checkUser();

    // Connect database
    ReconnectingGuard g(database_rc);

    // Get list of newsgroups
    StringList_t list;
    (DbHash(DbSubtree(database_connection, FORUM_ROOT).hashKey("newsgroups"))).getAll(list);

    // Build result
    Ptr<IntArrayData> ad(new IntArrayData());
    RenderState renderState(conn.getRenderState());
    renderState.setFormat("text");
    for (StringList_t::size_type i = 0, n = list.size(); i < n; i += 2) {
        string_t ng = list[i];
        int32_t fid;
        if (stringValue(list[i+1], fid)) {
            Forum f(database_connection, fid);
            if (conn.hasPrivilege(f.readPermissions().get())) {
                ad->content.pushNew(f.describeAsNewsgroup(renderState));
            }
        }
    }
    ad->addDimension(ad->content.size());
    IntArray a(ad);
    conn.sendObject(&a);
}

/* @q NNTPFINDNG name:Str (Talk Command)
   Find forum by newsgroup name.

   Permissions: read-access to forum.

   @err 404 Not found (name does not correspond to a forum)
   @err 403 Permission denied (user is not allowed to read this forum)
   @retval TalkNewsgroupInfo information
   @uses forum:newsgroups, forum:$FID:header */
void
doNntpFindNG(TalkConnection& conn, IntArgBlock& args)
{
    // Check args
    args.checkArgumentCount(1);
    string_t groupName = toStringResult(args.getNext());

    // Connect database
    ReconnectingGuard g(database_rc);

    // Look it up
    int32_t fid = DbHash(DbSubtree(database_connection, FORUM_ROOT).hashKey("newsgroups")).getIntField(groupName);
    if (fid == 0) {
        conn.sendError("404 Not found");
    } else {
        Forum f(database_connection, fid);
        RenderState renderState(conn.getRenderState());
        renderState.setFormat("text");
        if (conn.hasPrivilege(f.readPermissions().get())) {
            std::auto_ptr<IntValue> desc(f.describeAsNewsgroup(renderState));
            conn.sendObject(desc.get());
        } else {
            conn.sendError("403 Permission denied");
        }
    }
}

/* @q NNTPFINDMID mid:Str (Talk Command)
   Find posting by RFC Message-ID.
   The Message-ID is given without the angle brackets.

   Permissions: none.

   @err 404 Not found (Message-ID does not correspond to a posting)
   @retval MID posting Id */
void
doNntpFindMID(TalkConnection& conn, IntArgBlock& args)
{
    // Check args
    args.checkArgumentCount(1);
    string_t rfcMsgId = toStringResult(args.getNext());

    // Connect database
    ReconnectingGuard g(database_rc);

    // Look it up
    int32_t result = Message::lookupRfcMessageId(database_connection, rfcMsgId);
    if (result == 0) {
        conn.sendError("404 Not found");
    } else {
        IntIntIntValue iv(result);
        conn.sendObject(&iv);
    }
}

/* @q NNTPFORUMLS forum:FID (Talk Command)
   List forum.

   Permissions: none.

   @retval IntList list of message sequence numbers and posting Ids.
   The sequence numbers (even indexes) are in ascending order.
   @err 404 Not found */
void
doNntpForumList(TalkConnection& conn, IntArgBlock& args)
{
    // Check args
    args.checkArgumentCount(1);
    int32_t forumId = toIntResult(args.getNext());

    // Connect database
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, forumId);
    if (!f.exists()) {
        conn.sendError("404 Not found");
    }

    // Get result
    DbSort sorter(database_connection, f.messages().getName());
    sorter.by(MESSAGE_ROOT "*:header->seq");
    sorter.get(MESSAGE_ROOT "*:header->seq");
    sorter.get("#");

    std::auto_ptr<IntValue> result(sorter.getResult());
    conn.sendObject(result.get());
}

/* @q NNTPPOSTHEAD msg:MID (Talk Command)
   Get RFC message header for posting.
   The header contains regular RFC fields as well as some pseudo-headers.
   Field values are already preprocessed if necessary;
   The actual posting header can be produced by concatenating each field name, ": ", field value, "\r\n".

   Fields are returned in a useful order for generating the posting header
   and can thus be treated as a list (not a hash).
   The actual field selection is subject to change as needed
   and can possibly depend on the message.

   Permissions: user context required, read-access to message.

   @retkey :Id:MID (internal posting Id)
   @retkey :Seq:Int (sequence number)
   @retkey :Bytes:Int (estimate only)
   @retkey :Lines:Int (estimate only)
   @retkey Content-Transfer-Encoding:Str (synthetic)
   @retkey Content-Type:Str (synthetic)
   @retkey Date:Str (from message header)
   @retkey From:Str (depends on access checks)
   @retkey MIME-Version:Str (synthetic)
   @retkey Message-Id:Str (from message header or synthetic)
   @retkey Newsgroups:Str (synthetic)
   @retkey Path:Str (synthetic)
   @retkey Subject:Str (from message header)
   @retkey Supersedes:Str (from message header or synthetic)
   @retkey X-PCC-Posting-Id:MID (internal posting Id, for export)
   @retkey X-PCC-User:UserName (for export)
   @retkey Xref:Str (synthetic)

   @err 404 Message not found
   @see NNTPPOSTMHEAD
   @uses msg:$MID:header */
void
doNntpPostHead(TalkConnection& conn, IntArgBlock& args)
{
    // Check args
    args.checkArgumentCount(1);
    int32_t messageId = toIntResult(args.getNext());

    // Must have a user because getRfcHeader will return a user's email address
    conn.checkUser();

    // Connect database
    ReconnectingGuard g(database_rc);
    Message msg(database_connection, messageId);
    if (!msg.exists()) {
        conn.sendError("404 Message not found");
        return;
    }
    PostAccessChecker(conn).checkMessage(msg);

    // Return result
    std::auto_ptr<IntValue> head(msg.getRfcHeader());
    conn.sendObject(head.get());
}

/* @q NNTPPOSTMHEAD msg:MID... (Talk Command)
   Get RFC message header for multiple posting.
   For details, see {NNTPPOSTHEAD}.

   If one of the requested messages cannot be accessed,
   null is returned instead of the information; no error is generated.

   Permissions: none.

   @rettype List
   @uses msg:$MID:header
   @see NNTPPOSTHEAD */
void
doNntpPostMHead(TalkConnection& conn, IntArgBlock& args)
{
    // Must have a user because getRfcHeader will return a user's email address
    conn.checkUser();

    // Connect database
    ReconnectingGuard g(database_rc);

    // Build result
    IntDataSegment dseg;

    PostAccessChecker checker(conn);
    while (args.getNumArgs() > 0) {
        int32_t messageId = toIntResult(args.getNext());
        Message msg(database_connection, messageId);
        if (!msg.exists() || !checker.isAllowed(msg)) {
            dseg.pushNew(0);
        } else {
            dseg.pushNew(msg.getRfcHeader());
        }
    }
    conn.sendArray(dseg, dseg.size());
}

/* @q NNTPGROUPLS group:GRID (Talk Command)
   List forum group as newsgroup list.
   This is a variant of {GROUPLS}.

   Permissions: none.

   Everyone may list every group, but unlistable groups will appear empty in user contexts.
   Nonexistant groups will appear empty.

   The results will be sorted by their sort key.

   @retval StrList list of newsgroup names
   @see GROUPLS
   @uses group:$GRID:forums, forum:$FID:header */
void
doNntpGroupList(TalkConnection& conn, IntArgBlock& args)
{
    // Check args
    args.checkArgumentCount(1);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    Ptr<IntArrayData> ad(new IntArrayData());

    if (conn.isAdmin() || !group.unlisted().get()) {
        StringList_t result;
        group.forums().sort()
            .sortLexicographical()
            .by(FORUM_ROOT "*:header->key")
            .get(FORUM_ROOT "*:header->newsgroup")
            .getResult(result);

        for (StringList_t::const_iterator i = result.begin(); i != result.end(); ++i) {
            ad->content.pushNew(makeStringValue(*i));
        }
    }
    ad->addDimension(ad->content.size());

    // Send result
    IntArray arr(ad);
    conn.sendObject(&arr);
}
