/**
  *  \file planetscentral/talk/cmdthread.cc
  *  \brief Talk: THREAD commands
  */

#include "planetscentral/talk/cmdthread.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/dbclient/connection.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/cmdforum.h"
#include "planetscentral/talk/cmdpost.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/talkconnection.h"
#include "planetscentral/talk/topic.h"
#include "planetscentral/talk/forum.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/message.h"

/* @q THREADSTAT thread:TID (Talk Command)
   Get information about a forum thread.

   Permissions: read-access to thread.

   @retval TalkThreadInfo information about thread
   @err 404 Not found
   @uses thread:$TID:header */
void
doThreadStat(TalkConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    int32_t tid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, tid);
    if (!t.exists()) {
        conn.sendError("404 Not found");
        return;
    }
    PostAccessChecker(conn).checkTopic(t);
    std::auto_ptr<IntValue> result(t.describe());
    conn.sendObject(result.get());
}

/* @q THREADMSTAT thread:TID... (Talk Command)
   Get information about multiple forum threads.

   If one of the requested threads cannot be accessed,
   null is returned instead of the information; no error is generated.

   Permissions: none.

   @retval TalkThreadInfo[] information
   @uses thread:$TID:header */
void
doThreadMStat(TalkConnection& conn, IntArgBlock& args)
{
    // Do it
    ReconnectingGuard g(database_rc);
    IntDataSegment dseg;
    PostAccessChecker checker(conn);
    while (args.getNumArgs() > 0) {
        int32_t topicId = toIntResult(args.getNext());
        Topic t(database_connection, topicId);
        if (!t.exists() || !checker.isAllowed(t)) {
            dseg.pushNew(0);
        } else {
            dseg.pushNew(t.describe());
        }
    }
    conn.sendArray(dseg, dseg.size());
}

/* @q THREADLSPOST thread:TID [listParameters...] (Talk Command)
   List postings in a thread.

   The list can be accessed in different ways, see {pcc:talk:listparams|listParameters}.
   Valid sort keys for postings are:
   - author
   - edittime
   - subject
   - thread
   - time

   Permissions: none (everyone can execute this command).

   @rettype Any
   @rettype MID
   @uses thread:$TID:messages */
void
doThreadListPosts(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    ListParams parms;
    parms.parse(args, true, ListParams::postSortKeys);

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, parms.getId());
    if (!t.exists()) {
        conn.sendError("404 Not found");
    } else {
        parms.exec(conn, t.messages());
    }
}

/* @q THREADSTICKY thread:TID flag:Int (Talk Command)
   Set thread stickyness.
   We distinguish between sticky and normal (non-sticky) threads;
   sticky threads can be rendered separately.

   Permissions: delete-access to forum (treated as admin access).

   @err 404 Not found
   @uses thread:$TID:header, forum:$FID:threads, forum:$FID:stickythreads */
void
doThreadSticky(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    args.checkArgumentCount(2);
    int32_t topicId = toIntResult(args.getNext());
    int32_t value   = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, topicId);
    if (!t.exists()) {
        conn.sendError("404 Not found");
        return;
    }

    // Permission check
    conn.checkPrivilege(t.forum().deletePermissions().get());

    // Execute
    t.setSticky(value != 0);
    conn.sendSuccess("OK");
}

/* @q THREADPERMS thread:TID [perm:Str ...] (Talk Command)
   Get thread permissions.
   For each given permission name, checks whether the user has the respective privilege.
   - %read (read postings)
   - %write (create new threads)
   - %answer (answer to a posting)
   - %delete (delete postings)
   The returned value is an integer with each bit corresponding to a privilege.
   For example, "THREADPERMS 1 answer write" returns the "answer" permission in bit 0, the "write" permission in bit 1.

   If a permission is not set separately on the thread, the containing forum is consulted.

   Permissions: none (everyone can execute this command).

   @err 404 Not found
   @retval Int permissions
   @uses forum:$FID:header, thread:$TID:header */
void
doThreadPermissions(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    int32_t topicId = toIntResult(args.getNext());
    int32_t result = 0;
    int32_t mask = 1;

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, topicId);
    if (!t.exists()) {
        conn.sendError("404 Not found");
    } else {
        Forum f = t.forum();
        while (args.getNumArgs() > 0) {
            string_t key = toStringResult(args.getNext()) + "perm";
            string_t str = t.header().getStringField(key);
            if (str.empty()) {
                str = f.header().getStringField(key);
            }
            if (conn.hasPrivilege(str)) {
                result |= mask;
            }
            mask <<= 1;
        }
        IntIntIntValue iv(result);
        conn.sendObject(&iv);
    }
}

/* @q THREADMV thread:TID forum:FID (Talk Command)
   Move thread to another forum.
   This moves all postings within the thread.

   Permissions: delete-access to old forum, write-access to new forum.

   @err 404 Not found
   @uses thread:$TID:header, msg:$MID:header
   @uses forum:$FID:threads, forum:$FID:stickythreads, forum:$FID:messages */
void
doThreadMove(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t topicId = toIntResult(args.getNext());
    int32_t forumId = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, topicId);
    if (!t.exists()) {
        conn.sendError("404 Not found");
        return;
    }

    // Check forums
    Forum src(t.forum());
    Forum dst(database_connection, forumId);
    if (src.getId() == forumId) {
        conn.sendSuccess("OK (no change)");
        return;
    }

    // Check permissions: must have delete permission on source forum, and write permission on target
    if (!conn.isAdmin()) {
        conn.checkPrivilege(src.deletePermissions().get());
        conn.checkPrivilege(dst.writePermissions().get());
    }

    // Do it. Actual forum move is trivial, but we must update all sequence numbers and change
    // all message Ids for the NNTP side.
    IntList_t posts;
    t.messages().sort().by(MESSAGE_ROOT "*:header->seq").getResult(posts);
    for (uint32_t i = 0; i < posts.size(); ++i) {
        Message msg(database_connection, posts[i]);

        // NOTE: For a perfect mapping, we must actually make sure that the new sequence number does
        // not match ANY old sequence number, because otherwise the new message would have the same
        // generated message Id on the NNTP side. This could be trivially assured by choosing the
        // new one bigger than the old one, but that is not feasible (think moving a thread from a
        // 10000-post forum into a 10-post forum, leaving a 9990-post gap). Thus, we only avoid
        // clashes with the current and previous one.
        int32_t oldSeq1 = msg.sequenceNumber().get();
        int32_t oldSeq2 = msg.previousSequenceNumber().get();
        int32_t newSeq;
        do {
            newSeq = ++dst.lastMessageSequenceNumber();
        } while (newSeq == oldSeq1 || newSeq == oldSeq2);

        // Update the post's identifying information
        string_t rfcMsgId = msg.rfcMessageId().get();
        msg.removeRfcMessageId(database_connection, rfcMsgId);
        msg.previousRfcMessageId().set(rfcMsgId);
        msg.previousSequenceNumber().set(oldSeq1);
        msg.sequenceNumber().set(newSeq);
        msg.rfcMessageId().remove();
    }

    // Move the postings into the new forum
    src.messages().remove(t.messages()).storeTo(src.messages());
    dst.messages().merge(t.messages()).storeTo(dst.messages());

    // Move the thread
    if (t.isSticky()) {
        src.stickyTopics().moveTo(topicId, dst.stickyTopics());
    } else {
        src.topics().moveTo(topicId, dst.topics());
    }
    t.forumId().set(forumId);
    conn.sendSuccess("OK");
}

/* @q THREADRM thread:TID (Talk Command)
   Remove a thread.
   Removes all postings it contains.

   Permissions: delete-access to forum.

   @retval Int 0=thread did not exist, 1=thread removed
   @uses thread:$TID:header, msg:$MID:header
   @uses forum:$FID:threads, forum:$FID:stickythreads, forum:$FID:messages */
void
doThreadRemove(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    int32_t topicId = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, topicId);
    int32_t result;
    if (!t.exists()) {
        // Does not exist: assume success
        result = 0;
    } else {
        // Check delete permissions
        Forum f(t.forum());
        if (!conn.isAdmin()) {
            conn.checkPrivilege(f.deletePermissions().get());
        }

        // Do it
        t.remove();
        result = 1;
    }

    IntIntIntValue iv(result);
    conn.sendObject(&iv);
}
