/**
  *  \file planetscentral/talk/newsrc.cc
  *  \brief Talk: User newsrc access
  *
  *  user:$UID:forum:newsrc:data : hash
  *  user:$UID:forum:newsrc:index : int
  *          This represents a bitset.
  *          The bitset is split into lines of 1024 bytes = 8192 bits.
  *          Hash key is the line number, starting at 0.
  *          Each bit is 1 if the message is already read.
  *          Compaction:
  *          - treat all lines < index as "all-1"
  *          - otherwise, look into the hash
  *          - if line is missing, treat as "all-0"
  */

#include "planetscentral/talk/newsrc.h"

enum {
    LineShift = 13,
    LineSize  = 1 << LineShift,
    LineMask  = LineSize - 1,

    LineBytes = LineSize / 8
};


Newsrc::Newsrc(DbSubtree root)
    : root(root),
      readAllBelowLine(index().get()),
      cache(),
      cacheIndex(-1),
      cacheDirty(false)
{ }

void
Newsrc::save()
{
    // Remove entirely-read lines.
    while (cacheIndex == readAllBelowLine && cache.find_first_not_of(char(0xFF)) == string_t::npos) {
        // This line is entirely read, remove it.
        data().removeField(itoa(cacheIndex));
        ++readAllBelowLine;
        index().set(readAllBelowLine);

        // Load next one.
        doLoad(cacheIndex+1);
    }

    // If a dirty line remains, save it
    if (cacheDirty) {
        if (cache.find_first_not_of(char(0)) == string_t::npos) {
            data().removeField(itoa(cacheIndex));
        } else {
            data().setField(itoa(cacheIndex), cache);
            cacheDirty = false;
        }
    }
}

bool
Newsrc::get(int32_t messageId)
{
    // Split into line/column
    int32_t line = messageId >> LineShift;
    int32_t column = messageId & LineMask;

    // Quick handling?
    if (line < readAllBelowLine) {
        return true;
    }

    // Use cache
    loadCache(line);
    int32_t byte = column >> 3;
    int32_t bit = (column & 7);
    return ((uint8_t(cache[byte]) & (1 << bit)) != 0);
}

void
Newsrc::set(int32_t messageId)
{
    // Split into line/column
    int32_t line = messageId >> LineShift;
    int32_t column = messageId & LineMask;

    // Anything to do?
    if (line >= readAllBelowLine) {
        // Use cache
        loadCache(line);
        int32_t byte = column >> 3;
        int32_t bit = (column & 7);
        uint8_t mask = (1 << bit);
        if ((uint8_t(cache[byte]) & mask) == 0) {
            cache[byte] |= mask;
            cacheDirty = true;
        }
    }
}

void
Newsrc::clear(int32_t messageId)
{
    // Split into line/column
    int32_t line = messageId >> LineShift;
    int32_t column = messageId & LineMask;

    // Is the line we need actually available?
    while (line < readAllBelowLine) {
        // No, we have to create a new all-FF line
        save();
        --readAllBelowLine;
        cache.assign(std::size_t(LineSize), char(0xFF));
        cacheIndex = readAllBelowLine;
        cacheDirty = true;
        index().set(readAllBelowLine);
    }

    // Regular operation through cache
    loadCache(line);
    int32_t byte = column >> 3;
    int32_t bit = (column & 7);
    uint8_t mask = (1 << bit);
    if ((uint8_t(cache[byte]) & mask) != 0) {
        cache[byte] &= ~mask;
        cacheDirty = true;
    }
}

DbInt
Newsrc::index()
{
    return root.intKey("index");
}

DbHash
Newsrc::data()
{
    return root.hashKey("data");
}

void
Newsrc::loadCache(int32_t index)
{
    if (index != cacheIndex) {
        // Save old value
        save();
    }
    if (index != cacheIndex) {
        // Load new value. 'save' could have caused the new line to be loaded
        // already.
        doLoad(index);
    }
}

void
Newsrc::doLoad(int32_t index)
{
    cache      = data().getStringField(itoa(index));
    cacheIndex = index;
    cacheDirty = false;

    // Make cache value canonical
    if (cache.size() > LineBytes) {
        cache.erase(LineBytes);
    }
    if (cache.size() < LineBytes) {
        cache.append(LineBytes - cache.size(), char(0));
    }
}

#ifdef NEWSRC_TEST
/*
  g++-3.3 -DNEWSRC_TEST -O planetscentral/talk/newsrc.cc planetscentral/osutil.cc server/net.cc -o test_newsrc -I . -I cpluslib -L. -ldbclient -lpcc2 -lplatf -lcplus -lpthread
 */
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/dbclient/reconnect.h"
#include <stdio.h>

int main()
{
    // Connect
    Mutex output_mutex;
    DbReadyClient database_connection;
    ReconnectingConnection database_rc(database_connection, output_mutex, "127.0.0.1", "6666");
    database_rc.enter();

    // Test it
    Newsrc n(DbSubtree(database_connection, "test:"));

    char line[1000];
    while (fgets(line, sizeof(line), stdin)) {
        int a, b, c;
        if (strncmp(line, "sync", 4) == 0) {
            n.save();
        } else if (sscanf(line, "+%d-%d", &a, &b) == 2) {
            for (int i = a; i < b; ++i) {
                n.set(i);
            }
        } else if (sscanf(line, "+%d", &a) == 1) {
            n.set(a);
        } else if (sscanf(line, "-%d-%d", &a, &b) == 2) {
            for (int i = a; i < b; ++i) {
                n.clear(i);
            }
        } else if (sscanf(line, "-%d", &a) == 1) {
            n.clear(a);
        } else if (sscanf(line, "?%d-%d", &a, &b) == 2) {
            for (int i = a; i < b; ++i) {
                printf("%d: %d\n", i, n.get(i));
            }
        } else if (sscanf(line, "?%d", &a) == 1) {
            printf("%d: %d\n", a, n.get(a));
        } else {
            printf("usage: +RANGE (set), -RANGE (clear), ?RANGE (query), sync;\n"
                   "RANGE is index or interval A-B\n");
        }
    }
}
#endif
