/**
  *  \file planetscentral/talk/userpm.cc
  *  \brief Talk: User Message
  */

#include "planetscentral/talk/userpm.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/user.h"
#include "int/array.h"

/** Constructor.
    \param client Database client
    \param pmId Message Id */
UserPM::UserPM(DbClient& client, int32_t pmId)
    : pm(DbSubtree(client, PM_ROOT).subtree(pmId)),
      pmId(pmId)
{ }

/** Get message header. */
DbHash
UserPM::header()
{
    return pm.hashKey("header");
}

/** Message author (a user Id). */
DbHashStringField
UserPM::author()
{
    return header().stringField("author");
}

/** Message receivers.
    This can be a list of
    - u:uid (users)
    - g:gid (all players in a game)
    - g:gid:slot (slot in a game) */
DbHashStringField
UserPM::receivers()
{
    return header().stringField("to");
}

/** Message submission time. */
DbHashIntField
UserPM::time()
{
    return header().intField("time");
}

/** Message subject. */
DbHashStringField
UserPM::subject()
{
    return header().stringField("subject");
}

/** Message reference counter. */
DbHashIntField
UserPM::referenceCounter()
{
    return header().intField("ref");
}

/** Parent message identifier.
    0 if this is the initial message, otherwise pmid of the message this is a reply to. */
DbHashIntField
UserPM::parentMessageId()
{
    return header().intField("parent");
}

/** Message flags. */
DbHashIntField
UserPM::flags(string_t forUser)
{
    return header().intField("flags/" + forUser);
}

/** Message text. */
DbString
UserPM::text()
{
    return pm.stringKey("text");
}

/** Describe this message. */
IntValue*
UserPM::describe(string_t forUser)
{
    /* @type TalkMailInfo
       Information about one message.
       This is a variant of {pm:$PMID:header}.

       @key author:UID     (user who sent it)
       @key to:TalkAddr    (users who receive it)
       @key time:Time      (time when sent)
       @key subject:Str    (subject)
       @key flags:TalkFlag (user's flags)
       @key parent:PMID    (parent message, 0 if none) */
    Ptr<IntArrayData> ad(new IntArrayData());
    ad->content.pushNew(makeStringValue("author"));
    ad->content.pushNew(makeStringValue(author().get()));
    ad->content.pushNew(makeStringValue("to"));
    ad->content.pushNew(makeStringValue(receivers().get()));
    ad->content.pushNew(makeStringValue("time"));
    ad->content.pushNew(makeIntValue(time().get()));
    ad->content.pushNew(makeStringValue("subject"));
    ad->content.pushNew(makeStringValue(subject().get()));
    ad->content.pushNew(makeStringValue("flags"));
    ad->content.pushNew(makeIntValue(flags(forUser).get()));
    ad->content.pushNew(makeStringValue("parent"));
    ad->content.pushNew(makeIntValue(parentMessageId().get()));
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}

/** Get Id. */
int32_t
UserPM::getId() const
{
    return pmId;
}

/** Add a reference. */
void
UserPM::addReference()
{
    ++referenceCounter();
}

/** Remove a reference */
void
UserPM::removeReference()
{
    if (--referenceCounter() == 0) {
        header().remove();
        text().remove();
    }
}

/** Allocate a PM. */
int32_t
UserPM::allocatePM(DbClient& client)
{
    return ++DbSubtree(client, PM_ROOT).intKey("id");
}
